#ifndef RobotRunFileHelper_H
#define RobotRunFileHelper_H

#include <List>
#include <math.h>
#include <fstream.h>

//This class stores a single Pose, which is the cartesian coordinate (X,Y) position 
//and the angle a robot is facing.
class SosPose
{
public:
	SosPose operator=(const double num){_x = _y = _th = num; return *this;}
	
	SosPose(double x = 0, double y = 0, double th = 0)
    { 
		_x = x; 
		_y = y; 
		_th = th; 
	}
	
	SosPose(const SosPose &pose) : _x(pose._x), _y(pose._y), _th(pose._th) 
	{
	}	
	
	virtual ~SosPose() 
	{
	}
	
	virtual void setPose(double x, double y, double th = 0) 
    { 
		setX(x); 
		setY(y); 
		setTh(th); 
	}
	
	virtual void setPose(SosPose& otherPose)
    {
		setX(otherPose.getX());
		setY(otherPose.getY());
		setTh(otherPose.getTh());
    }
	
	void setX(double x) 
	{ 
		_x = x; 
	}
	
	void setY(double y) 
	{ 
		_y = y; 
	}
	
	void setTh(double th) 
	{ 
		_th = fixAngle(th); 
	}
	
	double getX(void) const 
	{ 
		return _x; 
	}
	
	double getY(void) const 
	{ 
		return _y; 
	}
	
	double getTh(void) const 
	{ 
		return _th; 
	}
	
	void getPose(double *x, double *y, double *th = NULL) 
    { 
		if (x != 0) 
		{
			*x = _x;
		}
		if (y != 0) 
		{
			*y = _y; 
		}
		if (th != 0) 
		{
			*th = _th; 
		}
    }
	
	virtual double findDistanceTo(SosPose otherPose)
    {
		return sqrt((getX() - otherPose.getX()) * (getX() - otherPose.getX()) +
			(getY() - otherPose.getY()) * (getY() - otherPose.getY()));
    }


	double radToDeg(double rad)  
	{ 
		return rad * 180.0 / 3.1415927; 
	}

	double fixAngle(double angle) 
	{
		if (angle >= 360)
		{
			angle = angle - 360.0 * (double)((int)angle / 360);
		}
		if (angle < -360)
		{
			angle = angle + 360.0 * (double)((int)angle / -360);
		}
		if (angle <= -180)
		{
			angle = + 180.0 + (angle + 180.0);
		}
		return angle;
	}
	
protected:
	double _x;
	double _y;
	double _th;
};

//The SensorReading class stores a single set of readings from range devices, either sonars or lasers.
//The constructor must be initialised with the number of range devices it will contain.
//This class stores both the Pose of the robot when these readings were taken, and the ranges of each
//of the sensors at that point in time.  The position of the sensors relative to the the pose of the
//robot is NOT stored, this is assumed to be stored somewhere else.
class SensorReading
{
public:
	//the default number of readings to store is 1
	SensorReading()
	{
		init(1);
	}

	//initialise the object with the number of sensor readings to be stored
	SensorReading(unsigned int numSensors)
	{
		init(numSensors);
	}

	SensorReading(const SensorReading& val)
	{
		_ranges = 0;
		*this = val;
	}

	~SensorReading()
	{
		delete [] _ranges;
	}

	//set the pose of the robot when the range readings were taken
	void setPose(SosPose pose)
	{
		_robotPose.setPose(pose);
	}

	//set the pose of the robot when the range readings were taken
	void setPose(double x, double y, double angle)
	{
		_robotPose.setPose(x,y,angle);
	}

	//get the pose of the robot when the range readings were taken
	const SosPose getPose() const
	{
		return _robotPose;
	}

	//set the range of sensor "rangeNum", to be the range "dist"
	//for example, if sensor 4 measured a range of 1500mm, you would call setRange(4,1500)
	void setRange(int rangeNum, long dist)
	{
		if(rangeNum < 0 || rangeNum >= _numSensors)
			return;

		_ranges[rangeNum] = dist;
	}

	//set the ranges for ALL sensors in one go.  The array "distances" is assumed to be at least
	//as big as the number of sensors that the object was initialised to.  If not, an error will occur.
	void setRange(long* distances)
	{
		for(int i = 0; i< _numSensors; i++)
		{
			_ranges[i] = distances[i];
		}
	}

	//get the range for the sensor numbered "num"
	const long getRange(int num) const 
	{
		if(num < 0 || num >= _numSensors)
			return 0;

		return _ranges[num];

	}

	//return the number of range sensors this object can store
	const int numSensors() const 
	{
		return (int) _numSensors;
	}

	//assignment operator
	virtual const SensorReading& operator=(const SensorReading& val)
	{
		
		if(_ranges != 0)
			delete [] _ranges;

		_robotPose = val.getPose();
		_numSensors = val.numSensors();

		_ranges = new long[_numSensors];

		for(int i = 0; i< _numSensors; i++)
		{
			_ranges[i] = val.getRange(i);
		}		
		
		return (*this);
	}

protected:
	unsigned int _numSensors;
	long* _ranges;
	SosPose _robotPose;

	//initialise the object
	void init(unsigned int numSensors)
	{
		if(numSensors <1)
			numSensors = 1;

		_numSensors = numSensors;
		_ranges = new long[_numSensors];

		for(int i = 0; i< _numSensors; i++)
		{
			_ranges[i]	= 0;
		}

		_robotPose.setPose(0,0,0);
	}	
};


//RobotRunFileHelper both reads and writes Robot Run files for the Map Viewer application, http://mapviewer.skynet.ie
//A RobotRun file contains the path that a robot took as a list of Poses, and optionally the sonar and/or laser
//range data that it read at each pose.  See the bottom of this file for an example of the format, or see the above 
//website for more information
class RobotRunFileHelper
{
public:
	//constructor
	RobotRunFileHelper()
	{
		reset();
	}

	//destructor
	~RobotRunFileHelper()
	{
		reset();
	}

	//reset the object - clear all lists and set all iterators to point to the start of all their corresponding lists
	void reset()
	{
		_laserReadingsList.clear();
		_sonarReadingsList.clear();
		_posesList.clear();
		_sonarPosesList.clear();
		_laserPosesList.clear();

		_sonarReadingsListIterator = _sonarReadingsList.begin();
		_laserReadingsListIterator = _laserReadingsList.begin();
		_posesListIterator = _posesList.begin();
		_sonarPosesListIterator = _sonarPosesList.begin();
		_laserPosesListIterator = _laserPosesList.begin();

		_robotRadius = 0;
	}

	//parses the file at the path "fileName"
	//Returns true if the file was semantically correct, i.e. if it was parsed correctly
	//Returns false if the file cannot be opened or if it is not correctly written
	bool parseFile(char* fileName)
	{
		reset();

		if(fileName == 0)
		{
			return false;
		}

		ifstream in;
		in.open(fileName);
		if(!in.is_open())
			return false;

		const int numEntries = 10;
		char validEntriesStatic[numEntries][20]={"numsonars","data","start","end","numlasers","robotradius",
												"sonar","laser"};

		char **validEntries = 0;
		int i = 0;//used in for loops

		validEntries = new char*[numEntries];
		for( i = 0; i< numEntries; i++)
		{
			validEntries[i] = validEntriesStatic[i];
		}

		const int TOKEN_NUMSONAR = 0, TOKEN_DATA = 1, TOKEN_START = 2,TOKEN_END = 3,
			TOKEN_NUMLASER = 4, TOKEN_ROBOTRADIUS = 5, TOKEN_SONAR = 6, TOKEN_LASER = 7;

		const int STATE_SONARLIST = 0, STATE_LASERLIST = 2,STATE_DATA = 3;

		const int BASE_STATE = -1;
		int currentState = BASE_STATE;
		char buffer[250];
		bool fileIsBad = false;
		int entryType = -1;

		in>>buffer;

		if(stricmp(buffer,"robotrun") != 0)
		{
			fileIsBad = true;
		}

		int numSonars = -1, numLasers = -1, numSensorsRead = 0, numSensors = 0;

		std::list<SensorReading> * sensorList = 0;

		SosPose pose(0,0,0);
		double * data = 0;
		SensorReading* sensorReading = 0;

		bool readSonarLayout = false, readLaserLayout = false;

		in>>buffer;

		while(!in.eof() && !fileIsBad)
		{
			switch(currentState)
			{
			case BASE_STATE:
				entryType = stringInArrayNoCase(buffer,validEntries,numEntries);

				switch(entryType)
				{
				case TOKEN_NUMSONAR:
					in>>buffer;
					if(!is_integer(buffer))
					{
						fileIsBad = true;
						break;
					}
					numSonars = atol(buffer);
					break;
				case TOKEN_DATA :
					in>>buffer;
					if(stricmp(buffer,"sonar") == 0 && numSonars > 0 && readSonarLayout)
					{
						sensorList = &_sonarReadingsList;
						numSensors = numSonars;
					}
					else if(stricmp(buffer,"laser") == 0 && numLasers > 0 && readLaserLayout)
					{
						sensorList = &_laserReadingsList;
						numSensors = numLasers;
					}
					else if(stricmp(buffer,"pose") == 0)
					{						
						sensorList = &_posesList;
						numSensors = 0;						
					}
					else
					{
						fileIsBad = true;	
						break;
					}

					currentState = STATE_DATA;
					sensorReading = new SensorReading(numSensors);
					
					data = new double[numSensors + 3];
					break;
				case TOKEN_START:
					in>>buffer;
					if(stricmp(buffer,"sonar") == 0 && numSonars != 0)
					{
						currentState = STATE_SONARLIST;
						_sonarPosesList.clear();
						numSensorsRead = 0;
						numSensors = numSonars;
					}
					else if(stricmp(buffer,"laser") == 0 && numLasers != 0)
					{
						currentState = STATE_LASERLIST;
						_laserPosesList.clear();
						numSensorsRead = 0;
						numSensors = numLasers;
					}
					else
					{
						fileIsBad = true;					
					}
					break;
				case TOKEN_NUMLASER :
					in>>buffer;
					if(!is_integer(buffer))
					{
						fileIsBad = true;
						break;
					}
					numLasers = atol(buffer);
					break;
				case TOKEN_ROBOTRADIUS :
					in>>buffer;
					if(!is_integer(buffer))
					{
						fileIsBad = true;
						break;
					}
					_robotRadius = atol(buffer);
					break;
				default:
					fileIsBad = true;
					break;
				}
				break;
			case STATE_SONARLIST:
				if(is_numeric(buffer) && numSensorsRead <= numSensors)
				{
					pose.setX(atol(buffer));

					in>>buffer;
					if(!is_numeric(buffer))
					{
						fileIsBad = true;
						break;
					}
					pose.setY(atol(buffer));

					in>>buffer;
					if(!is_numeric(buffer))
					{
						fileIsBad = true;
						break;
					}
					pose.setTh(atol(buffer));

					_sonarPosesList.push_back(pose);
					numSensorsRead++;
				}
				else if(stricmp(buffer, "end") == 0)
				{
					currentState = BASE_STATE;
					numSensorsRead = numSensors = 0;
					readSonarLayout = true;
				}
				else
				{
					fileIsBad = true;
				}

				break;
			case STATE_LASERLIST:
				if(is_numeric(buffer) && numSensorsRead <= numSensors)
				{
					pose.setX(atol(buffer));

					in>>buffer;
					if(!is_numeric(buffer))
					{
						fileIsBad = true;
						break;
					}
					pose.setY(atol(buffer));

					in>>buffer;
					if(!is_numeric(buffer))
					{
						fileIsBad = true;
						break;
					}
					pose.setTh(atol(buffer));

					_laserPosesList.push_back(pose);
					numSensorsRead++;
				}
				else if(stricmp(buffer, "end") == 0)
				{
					currentState = BASE_STATE;
					numSensorsRead = numSensors = 0;
					readLaserLayout = true;
				}
				else
				{
					fileIsBad = true;
				}

				break;
			case STATE_DATA:
				if(sensorList == 0 || data == 0 ||sensorReading == 0)
				{
					//this would be more of a parser error-cannot get here without the file being correct
					fileIsBad = true;
					break;
				}

				if(is_numeric(buffer))
				{
					data[0] = atof(buffer);

					for(i = 1; i < numSensors+3; i++)
					{
						in>>buffer;
						if(!is_numeric(buffer))
						{
							fileIsBad = true;
							break;
						}
						data[i] = atof(buffer);
					}

					sensorReading->setPose(data[0],data[1],data[2]);
					for(i = 0; i<numSensors; i++)
					{
						sensorReading->setRange(i,(long)data[i+3]);
					}

					sensorList->push_back(*sensorReading);
				}
				else if(stricmp(buffer,"end") == 0)
				{
					currentState = BASE_STATE;
					
					sensorList = 0;
					delete []data;
					data = 0;
					delete sensorReading;
					sensorReading = 0;
				}
				else
				{
					fileIsBad = true;
					break;
				}
				break;
			default:
				fileIsBad = true;
				break;

			}

			in>>buffer;
		}

		if(fileIsBad)
		{
			reset();
			in.close();

			return false;
		}

		return true;
	}

	//return the number of sonars on the robot
	int numSonars()
	{
		return _sonarPosesList.size();
	}

	//return the number of lasers on the robot
	int numLasers()
	{
		return _laserPosesList.size();
	}

	//return the number of sonar range readings in the file
	long numSonarReadings()
	{
		return _sonarReadingsList.size();
	}

	//return the number of laser range readings in the file
	long numLaserReadings()
	{
		return _laserReadingsList.size();
	}

	//return the number of pose readings in the file
	long numPoseReadings()
	{
		return _posesList.size();
	}
	
//-----------------------------------
	//reset the iterator for the Sonar range readings list
	void resetSonarReadingIterator()
	{
		_sonarReadingsListIterator = _sonarReadingsList.begin();
	}
	
	//reset the iterator for the Laser range readings list
	void resetLaserReadingIterator()
	{
		_laserReadingsListIterator = _laserReadingsList.begin();
	}

	//reset the iterator for the Pose readings list
	void resetPosesIterator()
	{
		_posesListIterator = _posesList.begin();
	}

//----------------------------------------

	//reset the iterator for the Sonar Pose list.  This list contains the position of each sonar in relation
	//to the direction the robot is facing, and also it's placement relative to the centre of the robot.
	void resetSonarPoseIterator()
	{	
		_sonarPosesListIterator = _sonarPosesList.begin();
	}

	//reset the iterator for the Laser Pose list.  This list contains the position of each laser in relation
	//to the direction the robot is facing, and also it's placement relative to the centre of the robot.
	void resetLaserPoseIterator()
	{	
		_laserPosesListIterator = _laserPosesList.begin();
	}
//----------------------------------------
	//get the next Sonar Range Reading.  Returns false if it has reached the end of the list 
	//(and therefore a valid reading was NOT returned), otherwise returns true
	bool getNextSonarReading(SensorReading& reading)
	{		
		if(_sonarReadingsListIterator == _sonarReadingsList.end())
		{
			return false;
		}

		reading = *_sonarReadingsListIterator;
		_sonarReadingsListIterator++;
		return true;
	}

	//get the next Sonar Range Reading, and remove it from the list.  Can save memory if you only need to 
	//read the reading and discard it.
	//Returns false if it has reached the end of the list 
	//(and therefore a valid reading was NOT returned), otherwise returns true
	bool popSonarReading(SensorReading& reading)
	{		
		if(_sonarReadingsList.size() > 0)
		{
			reading = _sonarReadingsList.front();

			_sonarReadingsList.pop_front();
			return true;
		}
		else
		{
			return false;
		}
	}
//----------------------------------------------
	//get the next Laser Range Reading.  Returns false if it has reached the end of the list 
	//(and therefore a valid reading was NOT returned), otherwise returns true
	bool getNextLaserReading(SensorReading& reading)
	{
		
		if(_laserReadingsListIterator == _laserReadingsList.end())
		{
			return false;
		}

		reading = *_laserReadingsListIterator;
		_laserReadingsListIterator++;
		return true;
	}

	//get the next Laser Range Reading, and remove it from the list.  Can save memory if you only need to 
	//read the reading and discard it.
	//Returns false if it has reached the end of the list 
	//(and therefore a valid reading was NOT returned), otherwise returns true
	bool popLaserReading(SensorReading& reading)
	{
		
		if(_laserReadingsList.size() > 0)
		{
			reading = _laserReadingsList.front();
			_laserReadingsList.pop_front();
			return true;
		}
		else
		{
			return false;
		}
	}

	//----------------------------------------
	
	//returns the pose of a sonar range device in relation to the centre of the robot
	bool getNextSonarPose(SosPose& pose)
	{
		
		if(_sonarPosesListIterator == _sonarPosesList.end())
		{
			return false;
		}

		pose = *_sonarPosesListIterator;
		_sonarPosesListIterator++;
		return true;
	}

	//returns the pose of a laser range device in relation to the centre of the robot
	bool getNextLaserPose(SosPose& pose)
	{
		
		if(_laserPosesListIterator == _laserPosesList.end())
		{
			return false;
		}

		pose = *_laserPosesListIterator;
		_laserPosesListIterator++;
		return true;
	}

//----------------------------------------------

	//get the next Pose Reading.  Pose readings are where there is no range data, as can be seen in the 
	//"DATA POSE" section in the example at the bottom of the file.
	//Returns false if it has reached the end of the list 
	//(and therefore a valid reading was NOT returned), otherwise returns true
	bool getNextPose(SensorReading& reading)
	{
		
		if(_posesListIterator == _posesList.end())
		{
			return false;
		}

		reading = *_posesListIterator;
		_posesListIterator++;
		return true;
	}
//----------------------------------------------

	//gets the radius of the robot
	long getRobotRadius()
	{
		return _robotRadius;
	}

	//set the robot radius
	void setRobotRadius(long radius)
	{
		_robotRadius = radius;
	}

//----------------------------------------------

	//push a sonar range reading onto the list
	void pushSonarReading(const SensorReading & reading)
	{
		_sonarReadingsList.push_back(reading);
	}

	//push a laser range reading onto the list
	void pushLaserReading(const SensorReading & reading)
	{
		_laserReadingsList.push_back(reading);
	}

	//Push the pose of a sonar onto the list
	void pushSonarPose(const SosPose & pose)
	{
		_sonarPosesList.push_back(pose);
	}

	//push the pose of a laser onto the list
	void pushLaserPose(const SosPose & pose)
	{
		_laserPosesList.push_back(pose);
	}

	//push a pose reading onto the list
	void pushPoseReading(const SensorReading& reading)
	{
		_posesList.push_back(reading);
	}

//--------------------------------------------------

	//write the information stored in the lists and robot radius to a the file "fileName".
	//For sonar data to be written, there must be information both on the positions of the sonars 
	//using the pushSonarPose method, and on the range data from the sonars using the pushSonarReading method.
	//The same goes for laser data.
	//For just Pose data, only Pose data has to be entered, using the pushPoseReading method
	bool writeToFile(char* fileName)
	{
		//if an invalid file name is passed in, return false
		if(fileName == 0)
			return false;

		ofstream out;
		out.open(fileName);

		//if we can't open the file, return false
		if(!out.is_open())
			return false;

		out<<"robotrun\n";
		SosPose pose;
		SensorReading reading;
		int numSensors = -1;

		out<<"robotradius "<<_robotRadius<<endl;

		//if there is sonar information on both the layout of the sonars and
		//data collected from the sonars, then write them to the file
		if(_sonarPosesList.size() > 0 && _sonarReadingsList.size() > 0)
		{
			resetSonarPoseIterator();
			numSensors = _sonarPosesList.size();
			out<<"numsonars "<<numSensors;
			out<<"\nstart sonar";

			while(getNextSonarPose(pose))
			{
				out<<"\n"<<pose.getX()<<' '<<pose.getY()<<' '<<pose.getTh();
			}
			out<<"\nend";

			out<<"\nDATA SONAR";
			resetSonarReadingIterator();
			while(getNextSonarReading(reading))
			{
				pose = reading.getPose();

				//output the pose of the robot when the readings were taken
				out<<"\n"<<pose.getX()<<' '<<pose.getY()<<' '<<pose.getTh();
				
				//output the range reading of the sonars
				for(int i = 0; i< numSensors; i++)
				{
					out<<' '<<reading.getRange(i);
				}
			}
			out<<"\nEND";
		}

		//if there is laser information on both the layout of the lasers and
		//data collected from the lasers, then write them to the file
		if(_laserPosesList.size() > 0 && _laserReadingsList.size() > 0)
		{
			resetLaserPoseIterator();
			numSensors = _laserPosesList.size();
			out<<"numlasers "<<numSensors;
			out<<"\nstart laser";

			while(getNextLaserPose(pose))
			{
				out<<"\n"<<pose.getX()<<' '<<pose.getY()<<' '<<pose.getTh();
			}
			out<<"\nend";

			out<<"\nDATA LASER";
			resetSonarReadingIterator();
			while(getNextLaserReading(reading))
			{
				pose = reading.getPose();

				//output the pose of the robot when the readings were taken
				out<<"\n"<<pose.getX()<<' '<<pose.getY()<<' '<<pose.getTh();
				
				//output the range reading of the sonars
				for(int i = 0; i< numSensors; i++)
				{
					out<<' '<<reading.getRange(i);
				}
			}
			out<<"\nEND";
		}

		if(_posesList.size() > 0)
		{
			resetPosesIterator();

			out<<"\nDATA POSE";

			while(getNextPose(reading))
			{
				pose = reading.getPose();
				out<<endl<<pose.getX()<<' '<<pose.getY()<<' '<<pose.getTh();
			}
			out<<"\nEND";

		}


		return true;
	}


private:
	std::list<SensorReading>		_sonarReadingsList;
	std::list<SensorReading>		_laserReadingsList;
	std::list<SensorReading>		_posesList;
	std::list<SosPose>				_sonarPosesList;
	std::list<SosPose>				_laserPosesList;

	std::list<SensorReading>::const_iterator	_sonarReadingsListIterator;
	std::list<SensorReading>::const_iterator	_laserReadingsListIterator;
	std::list<SensorReading>::const_iterator	_posesListIterator;
	std::list<SosPose>::const_iterator			_sonarPosesListIterator;
	std::list<SosPose>::const_iterator			_laserPosesListIterator;
	
	long							_robotRadius;

	int stringInArrayNoCase(char* str, char** strArray, int arrSize)
	{
		if(str == 0 || strlen(str) == 0 || strArray == 0 || arrSize < 1)
			return -1;

		for(int i = 0; i < arrSize; i++)
		{
			if(stricmp(str,strArray[i]) == 0)
				return i;
		}

		return -1;
	}

	bool is_integer(char * str)
	{
		if ( strspn ( str, "0123456789+-" ) != strlen(str) ) 
			return ( false) ;
		else
			return ( true) ;
	}

	bool is_numeric(char * str) 
	{
		if ( strspn ( str, "0123456789.+-e" ) != strlen(str) ) 
			return ( false) ;
		else
			return ( true) ;
	}

};

/*Here is an example file with a few readings
robotrun

numsonars 7
start sonar
100 100 90
120 80 30
130 40 15
130 0 0
130 -40 -15
120 -80 -30
100 -100 -90
end

numlasers 7
start laser
100 100 90
120 80 30
130 40 15
130 0 0
130 -40 -15
120 -80 -30
100 -100 -90
end
robotradius 220
DATA SONAR
0 0 0 870 2349 2500 2500 2500 2500 926 
1 1 0 870 2349 2500 2500 2500 2500 926 
2 3 0 905 2349 2500 2500 2400 2500 894 
4 5 0 905 2500 2500 2500 2400 2500 894 
5 6 0 905 2500 2500 2500 2400 2500 894 
0 0 0 905 2500 2500 2500 2200 2500 895 
END
DATA LASER
0 0 0 870 2349 2500 2500 2500 2500 926 
1 1 0 870 2349 2500 2500 2500 2500 926 
2 3 0 905 2349 2500 2500 2400 2500 894 
4 5 0 905 2500 2500 2500 2400 2500 894 
5 6 0 905 2500 2500 2500 2400 2500 894 
0 0 0 905 2500 2500 2500 2200 2500 895 
END
DATA POSE
0 0 0 
1 1 0 
2 3 0 
4 5 0 
5 6 0 
0 0 0 
END
*/





#endif